import re
import tempfile
import zipfile
from pathlib import Path

from qbt_migrate.methods import backup_folder


def test_backup_folder():
    temp_dir = Path(tempfile.mkdtemp())  # Get a tmp dir
    fastresume_file_count = 20  # Number of test fastresume files to build
    torrent_file_count = 20  # Number of test torrent files to build
    invalid_file_count = 20  # Number of invalid files to build

    # Build test files
    for x in range(fastresume_file_count):
        with open(temp_dir / f'test_{x}.fastresume', 'w') as f:
            f.write(f'This is file {x}')

    for x in range(torrent_file_count):
        with open(temp_dir / f'test_{x}.torrent', 'w') as f:
            f.write(f'This is file {x}')

    for x in range(invalid_file_count):
        with open(temp_dir / f'test_{x}.invalid', 'w') as f:
            f.write(f'This is invalid file {x}')

    # 'test.zip' will fail the filename check in archive if it gets archived
    archive_path = temp_dir / 'test.zip'
    backup_folder(temp_dir, archive_path)  # Call backup_folder with torrent files included

    # Validate zip file
    with zipfile.ZipFile(archive_path, 'r') as archive:
        assert(archive.testzip() is None)  # Check no errors
        # Assert that it has the correct file count
        assert(len(archive.infolist()) == fastresume_file_count + torrent_file_count)
        for file in archive.infolist():
            # Get the integer of the file
            filename = re.findall(r'test_(\d+).(?:fastresume)?(?:torrent)?', file.filename)
            assert(len(filename) != 0)  # Ensure we got the integer
            filename = filename[0]
            assert(archive.read(file.filename) == bytes(f'This is file {filename}', 'utf-8'))  # Integrity Check

    archive_path = temp_dir / 'test2.zip'
    backup_folder(temp_dir, archive_path, False)

    # Validate zip file
    with zipfile.ZipFile(archive_path, 'r') as archive:
        assert (archive.testzip() is None)  # Check no errors
        # Assert that it has the correct file count
        assert (len(archive.infolist()) == fastresume_file_count)
        for file in archive.infolist():
            # Get the integer of the file
            filename = re.findall(r'test_(\d+).fastresume', file.filename)
            assert (len(filename) != 0)  # Ensure we got the integer
            filename = filename[0]
            assert (archive.read(file.filename) == bytes(f'This is file {filename}', 'utf-8'))  # Integrity Check



import logging
import argparse

from . import QBTBatchMove, discover_bt_backup_path


logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-e', '--existing-path', help='Existing root of path to look for.')
    parser.add_argument('-n', '--new-path', help='New root path to replace existing root path with.')
    parser.add_argument('-t', '--target-os', help='Target OS (converts slashes). '
                                                  'Default will auto-detect if conversion is needed '
                                                  'based on existing vs new.',
                        choices=['Windows', 'Linux', 'Mac'])
    parser.add_argument('-b', '--bt-backup-path', help='BT_Backup Path Override. '
                                                       f'Default is {discover_bt_backup_path()}')
    parser.add_argument('-s', '--skip-bad-files', help='Skips bad .fastresume files instead of exiting. '
                                                       'Default behavior is to exit.',
                        action='store_true', default=False)

    parser.add_argument('-l', '--log-level', help='Log Level, Default is INFO.',
                        choices=['DEBUG', 'INFO'], default='INFO')

    return parser.parse_args()


def main():
    args = parse_args()
    logging.basicConfig()
    logger.setLevel(args.log_level)
    logging.getLogger('qbt_migrate').setLevel(args.log_level)
    logging.getLogger('qbt_migrate').propagate = True
    qbm = QBTBatchMove()
    if args.bt_backup_path is not None:
        qbm.bt_backup_path = args.bt_backup_path
    else:
        bt_backup_path = input(f'BT_Backup Path {qbm.bt_backup_path}: ')
        if bt_backup_path.strip():
            qbm.bt_backup_path = bt_backup_path
    if args.existing_path is None:
        args.existing_path = input('Existing Path: ')
    if args.new_path is None:
        args.new_path = input('New Path: ')
    if args.target_os is None:
        args.target_os = input('Target OS (Windows, Linux, Mac, Blank for auto-detect): ')
    if args.target_os.strip() and args.target_os.strip().lower() not in ('windows', 'linux', 'mac'):
        raise ValueError(f'Target OS is not valid. Must be Windows, Linux, or Mac. Received:{args.target_os}')
    elif not args.target_os.strip():
        if '/' in args.existing_path and '\\' in args.new_path:
            logger.info('Auto detected target OS change. Will convert slashes to Windows.')
            args.target_os = 'windows'
        elif '\\' in args.existing_path and '/' in args.new_path:
            logger.info('Auto detected target OS change. Will convert slashes to Linux/Mac.')
            args.target_os = 'linux'
        else:
            args.target_os = None
    logger.debug(f'Existing Path: {args.existing_path}, New Path: {args.new_path}, '
                 f'Target OS: {args.target_os}, Skip Bad Files: {args.skip_bad_files}')
    qbm.run(args.existing_path, args.new_path, args.target_os, True, args.skip_bad_files)


if __name__ == '__main__':
    main()



import os
import logging
from threading import Thread
from datetime import datetime
from typing import Optional

import bencodepy
from bencodepy.exceptions import BencodeDecodeError

from .methods import discover_bt_backup_path, convert_slashes, backup_folder


logger = logging.getLogger(__name__)
bencode = bencodepy.Bencode(encoding="utf-8", encoding_fallback="all")


class QBTBatchMove(object):
    logger = logging.getLogger(__name__ + '.QBTBatchMove')

    def __init__(self, bt_backup_path: str = None):
        if bt_backup_path is None:
            bt_backup_path = discover_bt_backup_path()
        self.logger.debug(f'BT_backup Path: {bt_backup_path}')
        self.bt_backup_path = bt_backup_path
        self.discovered_files = None

    def run(self, existing_path: str, new_path: str, target_os: Optional[str] = None,
            create_backup: bool = True, skip_bad_files: bool = False):
        """
        Perform Batch Processing of path changes.
        :param existing_path: Existing path to look for
        :type existing_path: str
        :param new_path: New Path to replace with
        :type new_path: str
        :param target_os: If targeting a different OS than the source. Must be Windows, Linux, or Mac.
        :type target_os: str
        :param create_backup: Create a backup archive of the BT_Backup directory?
        :type create_backup: bool
        :param skip_bad_files: Skip .fastresume files that cannot be read successfully.
        :type skip_bad_files: bool
        """
        if not os.path.exists(self.bt_backup_path) or not os.path.isdir(self.bt_backup_path):
            raise NotADirectoryError(self.bt_backup_path)
        if create_backup:
            backup_filename = f'fastresume_backup{datetime.now().strftime("%Y%m%d%H%M%S")}.zip'
            backup_folder(self.bt_backup_path, os.path.join(self.bt_backup_path, backup_filename))

        self.logger.info(f'Searching for .fastresume files with path {existing_path} ...')
        for fast_resume in self.discover_relevant_fast_resume(self.bt_backup_path, existing_path, not skip_bad_files):
            # Fire and forget
            Thread(target=fast_resume.replace_paths, args=[existing_path, new_path,
                                                           target_os, True, False]).start()

    @classmethod
    def discover_relevant_fast_resume(cls, bt_backup_path: str, existing_path: str, raise_on_error: bool = True):
        """
        Find .fastresume files that contain the existing path.
        :param bt_backup_path: Path to BT_backup folder
        :type bt_backup_path: str
        :param existing_path: The existing path to look for
        :type existing_path: str
        :param raise_on_error: Raise if error parsing .fastresume files
        :type raise_on_error: bool
        :return: List of FastResume Objects
        :rtype: list[FastResume]
        """
        for file in os.listdir(bt_backup_path):
            if file.endswith('.fastresume'):
                try:
                    fast_resume = FastResume(os.path.join(bt_backup_path, file))
                except (
                    BencodeDecodeError,
                    FileNotFoundError,
                    ValueError
                ) as e:
                    if raise_on_error:
                        cls.logger.critical(f'Unable to parse {file}. Stopping Discovery!')
                        raise e
                    cls.logger.warning(f'Unable to parse {file}. Skipping!\n\n{e}')
                    continue
                if existing_path in fast_resume.save_path or \
                        existing_path in fast_resume.qbt_save_path:
                    yield fast_resume
        return

    @classmethod
    def backup_folder(cls, folder_path: str, archive_path: str):
        return backup_folder(folder_path, archive_path)

    @classmethod
    def update_fastresume(cls, fast_resume: 'FastResume', existing_path: str, new_path: str,
                          target_os: Optional[str] = None, save_file: bool = True, create_backup: bool = True):
        fast_resume.replace_paths(existing_path, new_path, target_os, save_file, create_backup)


class FastResume(object):
    logger = logging.getLogger(__name__ + '.FastResume')

    def __init__(self, file_path: str):
        self._file_path = os.path.realpath(file_path)
        if not os.path.exists(self.file_path) or not os.path.isfile(self.file_path):
            raise FileNotFoundError(self.file_path)
        self.logger.debug(f'Loading Fast Resume: {self.file_path}')
        self._data = bencode.read(self.file_path)
        self.logger.debug(f'Fast Resume ({self.file_path}) Init Complete.')

    @property
    def file_path(self) -> str:
        return self._file_path

    @property
    def backup_filename(self) -> str:
        return f'{self.file_path}.{datetime.now().strftime("%Y%m%d%H%M%S")}.bkup'

    @property
    def save_path(self) -> Optional[str]:
        if 'save_path' in self._data:
            return self._data['save_path']

    @property
    def qbt_save_path(self) -> Optional[str]:
        if 'qBt-savePath' in self._data:
            return self._data['qBt-savePath']

    @property
    def mapped_files(self):
        if 'mapped_files' in self._data:
            return self._data['mapped_files']

    def set_save_path(self, path: str, key: str = 'save_path', target_os: Optional[str] = None,
                      save_file: bool = True, create_backup: bool = True):
        if key not in ['save_path', 'qBt-savePath']:
            raise KeyError('When setting a save path, key must be `save_path` or `qBt-savePath`. '
                           f'Received {key}')
        if create_backup:
            self.save(self.backup_filename)
        if target_os is not None:
            path = convert_slashes(path, target_os)
        self.logger.debug(f'Setting {key}... Old: {self._data[key]}, New: {path}, Target OS: {target_os}')
        self._data[key] = path
        if save_file:
            self.save()

    def set_save_paths(self, path: Optional[str], qbt_path: Optional[str] = None, target_os: Optional[str] = None,
                       save_file: bool = True, create_backup: bool = True):
        if create_backup:
            self.save(self.backup_filename)
        if path:
            self.set_save_path(path, key='save_path', target_os=target_os, save_file=False, create_backup=False)
        qbt_path = path if qbt_path is None else qbt_path
        if qbt_path:
            self.set_save_path(qbt_path, key='qBt-savePath', target_os=target_os, save_file=False, create_backup=False)
        if self.mapped_files is not None and target_os is not None:
            self.logger.debug('Converting Slashes for mapped_files...')
            self._data['mapped_files'] = [convert_slashes(path, target_os) for path in self.mapped_files]
        if save_file:
            self.save()

    def save(self, file_name: Optional[str] = None):
        if file_name is None:
            file_name = self.file_path
        self.logger.info(f'Saving File {file_name}...')
        bencode.write(self._data, file_name)

    def replace_paths(self, existing_path: str, new_path: str, target_os: Optional[str] = None,
                      save_file: bool = True, create_backup: bool = True):
        self.logger.info(f'Replacing Paths in FastResume {self.file_path}...')
        new_save_path = self.save_path.replace(existing_path, new_path) if self.save_path is not None else None
        new_qbt_save_path = self.qbt_save_path.replace(existing_path, new_path) \
            if self.qbt_save_path is not None else None
        self.logger.debug(f'Existing Save Path: {existing_path}, New Save Path: {new_path}, '
                          f'Replaced Save Path: {new_save_path}')
        self.set_save_paths(path=str(new_save_path), qbt_path=new_qbt_save_path, target_os=target_os,
                            save_file=save_file, create_backup=create_backup)
        self.logger.info(f'FastResume ({self.file_path}) Paths Replaced!')



import os
import sys
import logging
import zipfile
from pathlib import Path
from typing import Union


logger = logging.getLogger(__name__)


def discover_bt_backup_path():
    logger.debug('Discovering BT_backup path...')
    if sys.platform.startswith('win32'):
        logger.debug('Windows System')
        return os.path.join(os.getenv('localappdata'), 'qBittorrent\\BT_backup')

    if os.path.isfile('/.dockerenv') and os.path.isdir('/config/qBittorrent/BT_backup'):
        # Default path for config under for image: https://docs.linuxserver.io/images/docker-qbittorrent
        logger.debug('qBittorrent Docker container detected')
        return '/config/qBittorrent/BT_backup'

    logger.debug('Linux/Mac System')
    return os.path.join(os.getenv('HOME'), '.local/share/data/qBittorrent/BT_backup')


def convert_slashes(path: str, target_os: str):
    if target_os.strip().lower() not in ('windows', 'linux', 'mac'):
        raise ValueError(f'Target OS is not valid. Must be Windows, Linux, or Mac. Received: {target_os}')
    if target_os.strip().lower() == 'windows':
        logger.debug('Convert to Windows Slashes')
        return path.replace('/', '\\')
    logger.debug('Convert to Unix Slashes')
    return path.replace('\\', '/')


def backup_folder(folder_path: Union[str, os.PathLike], archive_path: Union[str, os.PathLike],
                  include_torrents: bool = True):
    logger.info(f'Creating Archive {archive_path} ...')
    folder_path = Path(folder_path)
    archive_path = Path(archive_path)
    with zipfile.ZipFile(archive_path, 'w') as archive:
        for file in folder_path.iterdir():
            if file.name.endswith('.fastresume') or (include_torrents and file.name.endswith('.torrent')):
                logger.debug(f'Archiving {file} into {archive_path}...')
                archive.write(file)
    logger.info('Done!')
